/*-
 * Copyright (c) 2003, Patrick MARIE <mycroft@virgaria.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

/*
 * Divertd (Divert Daemon - mini IDS)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <pcap.h>
#include <sys/param.h>

#include <err.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/ip_fw.h>
#include <netinet/ip.h>

#include "common.h"
#include "divertd.h"

/* Globals */
typedef enum {
	cDebugMode,
	cDivertPort,
	cPcapLogFile,
	cAddDivertRule,
	cNextDivertRuleNumber,
	cLastDivertRuleNumber,
	cReInject,
	cAddRejectRule,
	cNextRejectRuleNumber,
	cLastRejectRuleNumber,
	cNumberPacketsForDivert,
	cNumberPacketsForReject,
	cTimeOutDivert,
	cTimeOutReject,
	cNoMoreOption
}               OpCodes;

configop        keywords[] = {
	{"DebugMode", cDebugMode, 0},
	{"DivertPort", cDivertPort, 1},
	{"PcapLogFile", cPcapLogFile, 1},
	{"AddDivertRule", cAddDivertRule, 0},
	{"NextDivertRuleNumber", cNextDivertRuleNumber, 1},
	{"LastDivertRuleNumber", cLastDivertRuleNumber, 1},
	{"ReInject", cReInject, 0},
	{"AddRejectRule", cAddRejectRule, 0},
	{"NextRejectRuleNumber", cNextRejectRuleNumber, 1},
	{"LastRejectRuleNumber", cLastRejectRuleNumber, 1},
	{"NumberPacketsForDivert", cNumberPacketsForDivert, 1},
	{"NumberPacketsForReject", cNumberPacketsForReject, 1},
	{"TimeOutDivert", cTimeOutDivert, 1},
	{"TimeOutReject", cTimeOutReject, 1},
	{NULL, 0, 0}
};

/* bg is background mode (1: daemon mode, 0: debug mode) */
unsigned int    bg;

/* Divert Port */
static unsigned short divert_port = 0;
static char    *divert_port_str = NULL;

/* Divert rules added */
static unsigned int add_divert_rule_to_ip = 0;
static unsigned int first_divert_rule_number = 0;
static unsigned int next_divert_rule_number = 0;
static unsigned int last_divert_rule_number = 0;

static unsigned int number_packets_divert = 0;
static unsigned int timeout_divert = 0;

/* Re-Inject packets */
static unsigned int reinject = 0;

/* Reject rules added */
static unsigned int add_reject_rule_to_ip = 0;
static unsigned int first_reject_rule_number = 0;
static unsigned int next_reject_rule_number = 0;
static unsigned int last_reject_rule_number = 0;

static unsigned int number_packets_reject = 0;
static unsigned int timeout_reject = 0;

/* Misc. vars */
static char    *pcaplogfile = NULL;
static time_t   lasttmcheck = 0;

/* Oops there is no more rule for rejecting/diverting packets */
static unsigned char no_reject_rule = 0;
static unsigned char no_divert_rule = 0;

/* Divert pcap file descriptors */
static pcap_t  *pcap_fd_p;
static pcap_dumper_t *pcap_dump_log;

/* pcap erreur buffer */
char            errbuf[PCAP_ERRBUF_SIZE];

/* confarg is fort checkoption */
static char    *confarg;
/* conffile is the configuration file */
static char    *conffile = CONFIG_FILE;

/* Static data is 0'ed at process startup */
static alist    hosts;

void
usage(void)
{
	printf("Divertd\n"
	       "by Patrick MARIE <mycroft@virgaria.org>\n\n"
	       "Command line options:\n"
	       " -c <configfile>     Change configuration file. Default is " CONFIG_FILE "\n"
	       "Nothing more yet.\n\n");
	return;
}

int
checkoption(const char *str, configop * list)
{
	char           *confargtmp;
	configop       *tmp = list;
	int             i = 0;

	do {
		if (!strncasecmp(tmp->name, str, strlen(tmp->name))) {
			confarg = NULL;
			if (tmp->arg) {
				/* Get config line arguments from a command. */
				confargtmp = strdup((char *)(str + strlen(tmp->name)));
				confarg = clean_string(confargtmp);
				free(confargtmp);

				if (!confarg)
					return (ESYNTAX);
			}
			return (tmp->opcode);
		}
		tmp = list + (++i);
	}
	while (tmp->name != NULL);

	return (EUNKNOW);
}

int
parse_options(int argc, char **argv)
{
	int             ch;
	char           *p;

	while ((ch = getopt(argc, argv, "h?c:")) != -1) {
		p = argv[optind - 1];
		switch (ch) {
		case 'c':
			conffile = strdup(p);
			break;

		case 'h':
		case '?':
		default:
			usage();
			exit(EXIT_SUCCESS);
			break;
		}
	}
	return 0;
}

int
load_conf_file(void)
{
	FILE           *file;
	char           *opt;
	char            line[2048];

	char           *ep;

	file = fopen(conffile, "r");
	if (file == NULL) {
		err(1, "file open");
		return -1;
	}
	while (fgets(line, sizeof(line), file)) {
		if (checkunvalidity(line))
			continue;

		if (!(opt = clean_string(line)))
			continue;

		switch (checkoption(opt, keywords)) {
		case EUNKNOW:
			log_warn("Config option '%s' unknown.\n", opt);
			break;
		case ESYNTAX:
			log_warn("Config option '%s' requires an argument.\n", opt);
			break;

		case cDebugMode:
			bg = 0;
			break;

		case cDivertPort:
			divert_port = strtol(confarg, &ep, 10);
			divert_port_str = strdup(confarg);
			break;

		case cPcapLogFile:
			pcaplogfile = strdup(confarg);
			break;

		case cAddDivertRule:
			add_divert_rule_to_ip = 1;
			break;

		case cReInject:
			reinject = 1;
			break;

		case cNextDivertRuleNumber:
			first_divert_rule_number = strtol(confarg, &ep, 10);
			next_divert_rule_number = first_divert_rule_number;
			break;

		case cLastDivertRuleNumber:
			last_divert_rule_number = strtol(confarg, &ep, 10);
			break;

		case cAddRejectRule:
			add_reject_rule_to_ip = 1;
			break;

		case cNextRejectRuleNumber:
			first_reject_rule_number = strtol(confarg, &ep, 10);
			next_reject_rule_number = first_reject_rule_number;
			break;

		case cLastRejectRuleNumber:
			last_reject_rule_number = strtol(confarg, &ep, 10);
			break;

		case cNumberPacketsForDivert:
			number_packets_divert = strtol(confarg, &ep, 10);
			break;

		case cNumberPacketsForReject:
			number_packets_reject = strtol(confarg, &ep, 10);
			break;

		case cTimeOutDivert:
			timeout_divert = strtol(confarg, &ep, 10);
			break;

		case cTimeOutReject:
			timeout_reject = strtol(confarg, &ep, 10);
			break;

		}
	}

	if (!last_divert_rule_number)
		last_divert_rule_number = first_divert_rule_number;

	if (!last_reject_rule_number)
		last_reject_rule_number = first_reject_rule_number;

	fclose(file);
	return 0;
}

int
reset_conf_vars(void)
{
	/* Am I in daemon mode (background) ? */
	/* Default is on */
	bg = 1;

	/* Divert Port */
	divert_port = 0;

	/* add a divert rule on a matchtng ip */
	add_divert_rule_to_ip = 0;

	/* add a reject rule on a matching ip */
	add_reject_rule_to_ip = 0;

	reinject = 0;

	first_divert_rule_number = 0;
	next_divert_rule_number = 0;
	last_divert_rule_number = 0;

	first_reject_rule_number = 0;
	next_reject_rule_number = 0;
	last_reject_rule_number = 0;

	number_packets_divert = DEFAULT_NUM_DIVERT;
	number_packets_reject = DEFAULT_NUM_REJECT;

	timeout_divert = DEFAULT_TO_DIVERT;
	timeout_reject = DEFAULT_TO_REJECT;

	return 0;
}


/* End of daemon (exiting) */
int
clean_conf_vars(void)
{
	if (conffile != CONFIG_FILE) {
		free(conffile);
		conffile = CONFIG_FILE;
	}
	if (divert_port_str != NULL) {
		free(divert_port_str);
		divert_port_str = NULL;
	}
	if (pcaplogfile != NULL) {
		free(pcaplogfile);
		pcaplogfile = NULL;
	}
	return 0;
}

struct _ahost  *
lookup_ahost(struct in_addr h)
{
	ahost          *t;

	if ((t = hosts.addrlist[HASH(h.s_addr)]) == NULL)
		return (NULL);

	while (t && t->addr.s_addr != h.s_addr) {
		t = t->next;
	}
	return (t);
}

struct _ahost  *
add_ahost(struct in_addr h)
{
	/*
	 * This function doesn't look for the existence in the list of the
	 * addr. So you should take care about using it well.
	 */

	/* Firstly, looks up for a empty case */

	int             i;
	struct _ahost  *new, *tmp;

	for (i = 0; i < LIST_SIZE; i++) {
		if (hosts.hastlist[i].addr.s_addr == 0)
			break;
	}

	/* XXX No empty box ;-(( */

	if (i == LIST_SIZE) {
		log_debug("Hash list: no free box for %s\n", inet_ntoa(h));
		return (NULL);
	}
	log_debug("Hash list: adding host %s\n", inet_ntoa(h));

	new = &hosts.hastlist[i];
	new->addr = h;
	new->counter = 0;
	tmp = hosts.addrlist[HASH(h.s_addr)];

	new->prev = NULL;

	if (tmp) {
		new->next = tmp;
		new->next->prev = new;
	} else
		new->next = NULL;

	hosts.addrlist[HASH(h.s_addr)] = new;

	return (new);
}

struct _ahost  *
inc_counter(struct in_addr h)
{
	struct _ahost  *r;

	if ((r = lookup_ahost(h)) == NULL)
		r = add_ahost(h);

	r->counter++;
	r->timestamp = time(NULL);

	return (r);
}

int
rem_ahost(struct in_addr h)
{
	ahost          *t, *ot = NULL;

	if ((t = hosts.addrlist[HASH(h.s_addr)]) == NULL)
		return (0);


	while (t && t->addr.s_addr != h.s_addr) {
		ot = t;
		t = t->next;
	}

	if (t) {
		if (ot) {
			ot->next = t->next;
			t->next->prev = ot;
		} else {
			hosts.addrlist[HASH(h.s_addr)] = t->next;
			t->next->prev = NULL;
		}

		bzero(t, sizeof(ahost));
	}
	return (0);
}

int
open_fds(void)
{
	char            save_log[MAXPATHLEN];

	if (!divert_port) {
		log_warn("Fatal: Divertd need a divert port.\n");
		clean_conf_vars();
		exit(EXIT_FAILURE);
	}
	pcap_fd_p = pcap_open_live(divert_port_str, 2048, 0, 0, errbuf);

	if (pcap_fd_p == NULL) {
		log_warn("pcap_open_live failed: %s\n", errbuf);
		clean_conf_vars();
		exit(EXIT_FAILURE);
	}
	log_info("Listening divert socket number '%d'\n", divert_port);

	if (pcaplogfile != NULL) {
		if (access(pcaplogfile, F_OK) == 0) {
			snprintf(save_log, MAXPATHLEN - 1, "%s.%i", pcaplogfile, time(NULL));
			if (rename(pcaplogfile, save_log) == -1) {
				log_warn("Can't backup %s to %s.\n", pcaplogfile, save_log);
			}
		}
		pcap_dump_log = pcap_dump_open(pcap_fd_p, pcaplogfile);
		if (pcap_dump_log == NULL) {
			log_warn("pcap_dump_open failed: %s", pcap_geterr(pcap_fd_p));
			free(pcaplogfile);
			pcaplogfile = NULL;
		}
	}
	return 0;
}

int
close_fds(void)
{
	if (pcap_fd_p != NULL) {
		pcap_close(pcap_fd_p);
		pcap_fd_p = NULL;
	}
	if (pcap_dump_log != NULL) {
		pcap_dump_close(pcap_dump_log);
		pcap_dump_log = NULL;
	}
	return 0;
}

int
do_dump(char *dumpme)
{
	FILE           *fd;
	char            buffer[1024];
	int             i;

	if (access(DUMP_FILE, F_OK) == 0)
		if (unlink(DUMP_FILE) == -1) {
			log_warn("Can't delete dump file.\n");
			return -1;
		}
	fd = fopen(DUMP_FILE, "w+");
	if (fd == NULL) {
		log_warn("Can't create dump file.\n");
		return -1;
	}
	for (i = 0; i < LIST_SIZE; i++) {
		if (hosts.hastlist[i].fwrule_div || hosts.hastlist[i].fwrule_rej)
			snprintf(buffer, sizeof(buffer) - 1, "%s:%i:%d:%d:%d\n",
				 inet_ntoa(hosts.hastlist[i].addr),
				 hosts.hastlist[i].timestamp,
				 hosts.hastlist[i].counter,
				 hosts.hastlist[i].fwrule_div,
				 hosts.hastlist[i].fwrule_rej);
	}

	fclose(fd);
	return 0;
}

int
load_dump(char *dumpyou)
{
	FILE           *fd;
	char            buffer[1024];
	int             ret;

	struct in_addr  t_addr;
	time_t          t_ts;
	struct _ahost  *tmp;
	int             t_counter, t_fwrule_div, t_fwrule_rej;

	fd = fopen(DUMP_FILE, "r");
	if (fd == NULL) {
		log_warn("Can't load dump file.\n");
		return -1;
	}
	while ((ret = fscanf(fd, "%s:%i:%d:%d:%d\n", buffer, &t_ts,
			  &t_counter, &t_fwrule_div, &t_fwrule_rej)) == 5) {
		t_addr.s_addr = inet_addr(buffer);

		if ((tmp = add_ahost(t_addr)) != NULL) {

			tmp->timestamp = t_ts;
			tmp->counter = t_counter;
			tmp->fwrule_div = t_fwrule_div;
			tmp->fwrule_rej = t_fwrule_rej;

		}
	}

	fclose(fd);

	return 0;
}

int
get_fw_rule(unsigned short fwrule)
{
	int             s;
	int             nalloc = 1024;
	struct ip_fw   *rd;
	int             nbytes;
	void           *data = malloc(nbytes);

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

	nbytes = nalloc;

	while (nbytes >= nalloc) {
		nalloc = nalloc * 2 + 200;
		nbytes = nalloc;
		if ((data = realloc(data, nbytes)) == NULL) {
			close(s);
			return -1;
		}
		if (getsockopt(s, IPPROTO_IP, IP_FW_GET, data, &nbytes) < 0) {
			free(data);
			close(s);
			return -1;
		}
	}

	for (rd = data;
	     rd->rulenum < 65535 && rd->rulenum != fwrule;
	     rd = (void *)rd + RULESIZE(rd));

	close(s);
	s = (rd->rulenum == fwrule);
	free(data);

	return s;
}

/* Returns fwrule on success, 0 on error */
int
add_fw_rule(struct in_addr ip_addr, int fwrule)
{
	int             s;
	u_int32_t       rulebuf[255];
	struct ip_fw   *rule = (struct ip_fw *) rulebuf;
	ipfw_insn      *cmd = (ipfw_insn *) rule->cmd;

	unsigned short  frs, fir;
	unsigned int   *n = NULL, f = 0, l = 0;
	unsigned char  *y = NULL;

	bzero(rulebuf, sizeof(rulebuf));

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

	if (s < 0) {
		log_warn("Raw socket error.\n");
		return (0);
	}
	if (fwrule == TDIVERT) {
		y = &no_divert_rule;
		f = first_divert_rule_number;
		n = &next_divert_rule_number;
		l = last_divert_rule_number;
	} else if (fwrule == TREJECT) {
		y = &no_reject_rule;
		f = first_reject_rule_number;
		n = &next_reject_rule_number;
		l = last_reject_rule_number;
	}
	log_warn("F: %d, N: %d, L: %d\n", f, *n, l);

	if (l == f) {
		if (!get_fw_rule(f))
			fir = f;
		else {
			log_warn("No rule useable for %s ;-(\n", inet_ntoa(ip_addr));
			*y = 1;
			return 0;
		}
	} else {
		frs = *n;
		do {
			fir = (*n)++;
			if (*n > l)
				*n = f;

			if (frs == *n) {
				log_warn("No rule useable for %s ;-( (2)\n", inet_ntoa(ip_addr));
				*y = 1;
				return 0;
			}
		} while (get_fw_rule(fir));
	}

	rule->rulenum = fir;

	log_debug("Using rule %d against %s\n", fir, inet_ntoa(ip_addr));

	cmd = fill_cmd(cmd, O_PROTO, F_INSN_SIZE(ipfw_insn), 0, IPPROTO_IP);
	cmd = fill_ip(cmd, O_IP_SRC, ip_addr.s_addr);

	rule->act_ofs = (u_int32_t *) cmd - (u_int32_t *) rule->cmd;
	if (fwrule == TDIVERT)
		cmd = fill_cmd(cmd, O_DIVERT, F_INSN_SIZE(ipfw_insn), 0, divert_port);
	if (fwrule == TREJECT)
		cmd = fill_cmd(cmd, O_REJECT, F_INSN_SIZE(ipfw_insn), 0, 0);
	rule->cmd_len = (u_int32_t *) cmd - (u_int32_t *) rule->cmd;

	if (setsockopt(s, IPPROTO_IP, IP_FW_ADD, rule,
		       (void *)cmd - (void *)rulebuf) == -1) {
		log_warn("Setsockopt error.\n");
		close(s);
		return 0;
	}
	close(s);
	return fir;
}

int
del_fw_rule(unsigned short fwrule)
{
	int             s;
	u_int32_t       rulenum = fwrule;

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

	if (setsockopt(s, IPPROTO_IP, IP_FW_DEL, &rulenum,
		       sizeof(rulenum)) == -1) {
		log_warn("Setsockopt error.\n");
		close(s);
		return -1;
	}
	close(s);
	return 0;
}

int
clean_up_firewall(void)
{
	int             i;
	for (i = 0; i < LIST_SIZE; i++) {
		if (hosts.hastlist[i].fwrule_div)
			del_fw_rule(hosts.hastlist[i].fwrule_div);
		if (hosts.hastlist[i].fwrule_rej)
			del_fw_rule(hosts.hastlist[i].fwrule_rej);
	}
	return 0;
}

void
sigfunc(int signal)
{
	switch (signal) {
		case SIGHUP:
		log_info("Reloading configuration.\n");
		reset_conf_vars();
		load_conf_file();
		close_fds();
		open_fds();
		break;
	case SIGINT:
		log_warn("SigInt recieved: exiting.\n");

		clean_up_firewall();
		close_fds();
		clean_conf_vars();
		exit(EXIT_SUCCESS);
		break;
	}
	return;
}

int
act_recv_packet(unsigned char *packet, struct pcap_pkthdr usefull)
{
	struct ip      *iphdr = (struct ip *) packet;
	struct in_addr  ip_addr;
	struct _ahost  *r;
	unsigned short  u;

	ip_addr = iphdr->ip_src;

	r = inc_counter(ip_addr);

	if (no_divert_rule)
		return 1;

	if (add_divert_rule_to_ip
	    && next_divert_rule_number
	    && r->counter >= number_packets_divert
	    && !r->fwrule_div) {

		if ((u = add_fw_rule(ip_addr, TDIVERT)) > 0) {
			log_info("(UT: %i) Adding a divert rule against %s\n",
				 r->timestamp, inet_ntoa(ip_addr));
			r->fwrule_div = u;

		}
	}
	if (no_reject_rule)
		return 1;

	if (add_reject_rule_to_ip
	    && next_reject_rule_number
	    && r->counter >= number_packets_reject) {
		if ((u = add_fw_rule(ip_addr, TREJECT)) > 0) {
			log_info("(UT: %i) Adding a REJECT rule against %s\n",
				 r->timestamp, inet_ntoa(ip_addr));
			r->fwrule_rej = u;
		}
	}
	return 0;
}

/*
 * int list_all(void) { struct _ahost  *t; int             i;
 * printf("beginning listing.\n"); for (i = 0; i < 1 << HASH_SIZE; i++) { if
 * (hosts.addrlist[i]) { t = hosts.addrlist[i]; while (t) { printf("hash: %x
 * - %x\n", hosts.addrlist[i], t->addr.s_addr); t = t->next; } } }
 * 
 * return 0; }
 */

int
check_timesout(void)
{
	struct _ahost  *t, *ot;
	unsigned int    current = time(NULL);
	int             i, remove_t;

	for (i = 0; i < (1 << HASH_SIZE); i++) {
		if (hosts.addrlist[i]) {
			t = hosts.addrlist[i];
			while (t) {
				remove_t = 0;
				ot = t->next;
				if (add_reject_rule_to_ip
				    && t->counter >= number_packets_reject) {
					if (current >= t->timestamp + timeout_reject) {
						if (t->fwrule_rej || t->fwrule_div) {
							log_info("(UT: %i) Removing rule for %s\n",
								 current, inet_ntoa(t->addr));
							if (t->fwrule_rej) {
								del_fw_rule(t->fwrule_rej);
								no_reject_rule = 0;
							}
							if (t->fwrule_div) {
								del_fw_rule(t->fwrule_div);
								no_divert_rule = 0;
							}
						}
						remove_t = 1;
					}
				} else if (add_divert_rule_to_ip
				   && t->counter >= number_packets_divert) {
					if (current >= t->timestamp + timeout_divert) {
						if (t->fwrule_div) {
							log_info("(UT: %i) Removing rule for %s\n",
								 current, inet_ntoa(t->addr));
							del_fw_rule(t->fwrule_div);
							no_divert_rule = 0;
						}
						remove_t = 1;
					}
				} else if (current >= t->timestamp + DEFAULT_TO_NORMAL) {
					remove_t = 1;
				}
				if (remove_t) {
					/* Ok I remove 't' */
					log_debug("Hash list: removing host %s\n",
						  inet_ntoa(t->addr));
					if (t->prev)
						t->prev->next = t->next;
					if (t->next)
						t->next->prev = t->prev;
					bzero(t, sizeof(ahost));
					hosts.addrlist[i] = NULL;
					t = ot;
				} else {
					t = t->next;
				}
			}
		}
	}

	return current;
}

int
main(int argc, char **argv)
{
	int             i;
	int             ret, max_fd;
	struct sigaction sig;
	fd_set          main_set, save_set;
	unsigned char  *packet;
	struct pcap_pkthdr usefull;
	struct timeval  ts;

	struct sockaddr sin;

	sig.sa_handler = (void *)sigfunc;
	sig.sa_flags = 0;

	sigaction(SIGHUP, &sig, NULL);
	sigaction(SIGINT, &sig, NULL);

	if (getuid())
		errx(1, "Permission'll be denied: You've to be root.");

	reset_conf_vars();

	parse_options(argc, argv);
	load_conf_file();

	if (bg) {
		daemon(0, 0);
	}
	open_fds();

	FD_ZERO(&main_set);
	FD_SET(pcap_fileno(pcap_fd_p), &main_set);
	max_fd = pcap_fileno(pcap_fd_p);

	save_set = main_set;
	ts.tv_sec = 3;
	ts.tv_usec = 0;

	lasttmcheck = time(NULL);

	while ((ret = select(1 + max_fd, &main_set, NULL, NULL, &ts)) != -1) {
		if (FD_ISSET(pcap_fileno(pcap_fd_p), &main_set)) {
			/* Packet */
			packet = (u_char *) pcap_next(pcap_fd_p, &usefull);

			if (reinject) {
				if (usefull.caplen != usefull.len) {
					log_warn("Will not reinject incomplete packet.\n");
					continue;
				}
				pcap_get_last_sockaddr(pcap_fd_p, &sin);

				i = sendto(pcap_fileno(pcap_fd_p), packet, usefull.len,
					   0, &sin, sizeof(struct sockaddr));
			}
			act_recv_packet(packet, usefull);

			if (pcap_dump_log != NULL) {
				pcap_dump((u_char *) pcap_dump_log, &usefull, packet);
			}
		}
		if ((add_reject_rule_to_ip || add_divert_rule_to_ip)
		    && lasttmcheck + 10 <= time(NULL)) {
			lasttmcheck = check_timesout();
		}
		main_set = save_set;
		ts.tv_sec = 3;
		ts.tv_usec = 0;
	}

	clean_up_firewall();
	close_fds();
	clean_conf_vars();

	return (EXIT_SUCCESS);
}
